#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include "dream_ffi.h"

Display *display = NULL;
XFontStruct *font_info;
int screen_num;
Window win;
Colormap default_cmap;
GC gc;

struct pair*
prim_x_open_window( struct pair* args )
{
    char* display_name;
    XEvent report;
    XGCValues values;
    unsigned int width, height;
    GET_NULLABLE_STRING( display_name, args );
    GET_INTEGER( width, args );
    GET_INTEGER( height, args );
    NO_MORE_ARGS( args );
    if( display ) return &dream_false;
    if ( (display = XOpenDisplay(display_name)) == NULL )
    {
        fprintf( stderr, "cannot connect to X server %s\n", 
                        XDisplayName(display_name));
        return &dream_false;
    }

    screen_num = DefaultScreen(display);
    default_cmap = DefaultColormap(display, screen_num);

    /* Load font and get font information structure. */
    if ((font_info = XLoadQueryFont(display, "9x15")) == NULL)
    {
        fprintf( stderr, "cannot open 9x15 font\n" );
        return &dream_false;
    }

    win = XCreateSimpleWindow(display, RootWindow(display,screen_num), 
                    0, 0, width, height, 4, BlackPixel(display, screen_num),
                    WhitePixel(display,screen_num));

    /* Select event types wanted */
    XSelectInput(display, win, ExposureMask );

    /* Create default Graphics Context */
    gc = XCreateGC(display, win, 0, &values);

    /* specify font */
    XSetFont(display, gc, font_info->fid);

    /* specify black foreground since default window background is 
     * white and default foreground is undefined. */
    XSetForeground(display, gc, BlackPixel(display,screen_num));

    /* Display window */
    XMapWindow(display, win);

    do {
        XNextEvent(display, &report);
    } while (report.type != Expose || report.xexpose.count != 0);

    return &dream_true;
}

struct pair*
prim_x_clear_window( struct pair* args )
{
    NO_MORE_ARGS( args );
    XClearWindow( display, win );
    return &dream_true;
}

struct pair*
prim_x_set_foreground( struct pair* args )
{
    XColor exact_def;
    char* color_name;
    GET_STRING(color_name, args);
    NO_MORE_ARGS( args );
    if( !XParseColor(display, default_cmap, color_name, &exact_def ) )
    {
        fprintf( stderr, "color name %s not in database\n", color_name );
        return &dream_false;
    }
    if( !XAllocColor(display, default_cmap, &exact_def) )
    {
        fprintf( stderr, "can't allocate color\n" );
        return &dream_false;
    }
    XSetForeground(display, gc, exact_def.pixel);
    return &dream_true;
}

struct pair*
prim_x_close_window( struct pair* args )
{
    NO_MORE_ARGS( args );
    if( !display ) return &dream_false;
    XUnloadFont(display, font_info->fid);
    XFreeGC(display, gc);
    XCloseDisplay(display);
    display = NULL;
    return &dream_true;
}

struct pair*
prim_x_flush( struct pair* args )
{
    NO_MORE_ARGS( args );
    XFlush( display );
    return &dream_true;
}

struct pair*
prim_x_draw_point( struct pair* args )
{
    int x, y;
    GET_INTEGER( x, args );
    GET_INTEGER( y, args );
    NO_MORE_ARGS( args );
    XDrawPoint( display, win, gc, x, y );
    return &dream_true;
}

struct pair*
prim_x_draw_line( struct pair* args )
{
    int x1, y1, x2, y2;
    GET_INTEGER( x1, args );
    GET_INTEGER( y1, args );
    GET_INTEGER( x2, args );
    GET_INTEGER( y2, args );
    NO_MORE_ARGS( args );
    XDrawLine( display, win, gc, x1, y1, x2, y2 );
    return &dream_true;
}

struct pair*
prim_x_draw_rectangle( struct pair* args )
{
    int x, y;
    unsigned int width, height;
    GET_INTEGER( x, args );
    GET_INTEGER( y, args );
    GET_INTEGER( width, args );
    GET_INTEGER( height, args );
    NO_MORE_ARGS( args );
    XDrawRectangle(display, win, gc, x, y, width, height);
    return &dream_true;
}

struct pair*
prim_x_fill_rectangle( struct pair* args )
{
    int x, y;
    unsigned int width, height;
    GET_INTEGER( x, args );
    GET_INTEGER( y, args );
    GET_INTEGER( width, args );
    GET_INTEGER( height, args );
    NO_MORE_ARGS( args );
    XFillRectangle(display, win, gc, x, y, width, height);
    return &dream_true;
}
